@@ -25,6 +25,7 @@
 from ctypes import *
 from ctypes.util import find_library
 import os
+
 # A short circuited or expression to find the FluidSynth library
 # (mostly needed for Windows distributions of libfluidsynth supplied with QSynth)
 
@@ -35,6 +36,7 @@ if hasattr(os, 'add_dll_directory'):
 
 lib = find_library('fluidsynth') or \
     find_library('libfluidsynth') or \
+    find_library('libfluidsynth-3') or \
     find_library('libfluidsynth-2') or \
     find_library('libfluidsynth-1')
     
@@ -59,13 +61,25 @@ def cfunc(name, result, *args):
         return None
 
 # Bump this up when changing the interface for users
-api_version = '1.3.0'
+api_version = '1.3.1'
 
 # Function prototypes for C versions of functions
 
 FLUID_OK = 0
 FLUID_FAILED = -1
 
+fluid_version = cfunc('fluid_version', c_void_p,
+                        ('major', POINTER(c_int), 1),
+                        ('minor', POINTER(c_int), 1),
+                        ('micro', POINTER(c_int), 1))
+
+majver = c_int()
+fluid_version(majver, c_int(), c_int())
+if majver.value > 1:
+    FLUIDSETTING_EXISTS = FLUID_OK
+else:
+    FLUIDSETTING_EXISTS = 1
+
 # fluid settings
 new_fluid_settings = cfunc('new_fluid_settings', c_void_p)
 
@@ -103,6 +117,34 @@ fluid_settings_getint = cfunc('fluid_settings_getint', c_int,
 delete_fluid_settings = cfunc('delete_fluid_settings', None,
                               ('settings', c_void_p, 1))
 
+fluid_synth_activate_key_tuning = cfunc('fluid_synth_activate_key_tuning', c_int,
+                                        ('synth', c_void_p, 1),
+                                        ('bank', c_int, 1),
+                                        ('prog', c_int, 1),
+                                        ('name', c_char_p, 1),
+                                        ('pitch', POINTER(c_double), 1),
+                                        ('apply', c_int, 1))
+
+fluid_synth_activate_tuning = cfunc('fluid_synth_activate_tuning', c_int,
+                                    ('synth', c_void_p, 1),
+                                    ('chan', c_int, 1),
+                                    ('bank', c_int, 1),
+                                    ('prog', c_int, 1),
+                                    ('apply', c_int, 1))
+
+fluid_synth_deactivate_tuning = cfunc('fluid_synth_deactivate_tuning', c_int,
+                                      ('synth', c_void_p, 1),
+                                      ('chan', c_int, 1),
+                                      ('apply', c_int, 1))
+
+fluid_synth_tuning_dump = cfunc('fluid_synth_tuning_dump', c_int,
+                                ('synth', c_void_p, 1),
+                                ('bank', c_int, 1),
+                                ('prog', c_int, 1),
+                                ('name', c_char_p, 1),
+                                ('length', c_int, 1),
+                                ('pitch', POINTER(c_double), 1))
+
 # fluid synth
 new_fluid_synth = cfunc('new_fluid_synth', c_void_p,
                         ('settings', c_void_p, 1))
@@ -302,7 +344,7 @@ fluid_synth_set_midi_router = cfunc('fluid_synth_set_midi_router', None,
                                ('synth', c_void_p, 1),
                                ('router', c_void_p, 1))
 
-fluid_synth_handle_midi_event = cfunc('fluid_synth_handle_midi_event', POINTER(c_int),
+fluid_synth_handle_midi_event = cfunc('fluid_synth_handle_midi_event', c_int,
                                ('data', c_void_p, 1),
                                ('event', c_void_p, 1))
 
@@ -377,10 +419,68 @@ fluid_event_noteoff = cfunc('fluid_event_noteoff', None,
                          ('channel', c_int, 1),
                          ('key', c_short, 1))
 
-
 delete_fluid_event = cfunc('delete_fluid_event', None,
                           ('evt', c_void_p, 1))
 
+fluid_midi_event_get_channel = cfunc('fluid_midi_event_get_channel', c_int,
+                                  ('evt', c_void_p, 1))
+
+fluid_midi_event_get_control = cfunc('fluid_midi_event_get_control', c_int,
+                                  ('evt', c_void_p, 1))
+
+fluid_midi_event_get_program = cfunc('fluid_midi_event_get_program', c_int,
+                                  ('evt', c_void_p, 1))	
+
+fluid_midi_event_get_key = cfunc('fluid_midi_event_get_key', c_int,
+                                  ('evt', c_void_p, 1))
+
+fluid_midi_event_get_type = cfunc('fluid_midi_event_get_type', c_int,
+                                  ('evt', c_void_p, 1))
+
+fluid_midi_event_get_value = cfunc('fluid_midi_event_get_value', c_int,
+                                  ('evt', c_void_p, 1))
+
+fluid_midi_event_get_velocity = cfunc('fluid_midi_event_get_velocity', c_int,
+                                  ('evt', c_void_p, 1))
+
+# tempo_type used by fluid_player_set_tempo()
+FLUID_PLAYER_TEMPO_INTERNAL = 0
+FLUID_PLAYER_TEMPO_EXTERNAL_BPM = 1
+FLUID_PLAYER_TEMPO_EXTERNAL_MIDI = 2
+
+new_fluid_player = cfunc('new_fluid_player', c_void_p,
+                          ('synth', c_void_p, 1))
+
+delete_fluid_player = cfunc('delete_fluid_player', None,
+                             ('player', c_void_p, 1))
+
+fluid_player_add = cfunc('fluid_player_add', c_void_p,
+                         ('player', c_void_p, 1),
+                         ('filename', c_char_p, 1))
+
+fluid_player_join = cfunc('fluid_player_join', c_int,
+                          ('player', c_void_p, 1))
+
+fluid_player_play = cfunc('fluid_player_play', c_int,
+                          ('player', c_void_p, 1))
+
+fluid_player_set_playback_callback = cfunc('fluid_player_set_playback_callback', c_int,
+                                           ('player', c_void_p, 1),
+                                           ('handler', CFUNCTYPE(c_int, c_void_p, c_void_p), 1),
+                                           ('event_handler_data', c_void_p, 1))
+
+fluid_player_set_tempo = cfunc('fluid_player_set_tempo', c_int,
+                               ('player', c_void_p, 1),
+                               ('tempo_type', c_int, 1),
+                               ('tempo', c_double, 1))
+
+fluid_player_seek = cfunc('fluid_player_seek', c_int,
+                          ('player', c_void_p, 1),
+                          ('ticks', c_int, 1))
+
+fluid_player_stop = cfunc('fluid_player_stop', c_int,
+                          ('player', c_void_p, 1))
+
 # fluid audio driver
 new_fluid_audio_driver = cfunc('new_fluid_audio_driver', c_void_p,
                                ('settings', c_void_p, 1),
@@ -392,7 +492,7 @@ delete_fluid_audio_driver = cfunc('delete_fluid_audio_driver', None,
 # fluid midi driver
 new_fluid_midi_driver = cfunc('new_fluid_midi_driver', c_void_p,
                                ('settings', c_void_p, 1),
-                               ('handler', CFUNCTYPE(POINTER(c_int), c_void_p, c_void_p), 1),
+                               ('handler', CFUNCTYPE(c_int, c_void_p, c_void_p), 1),
                                ('event_handler_data', c_void_p, 1))
 
 
@@ -438,10 +538,10 @@ fluid_midi_router_rule_set_param2 = cfunc('fluid_midi_router_rule_set_param2', N
 # fluid midi router
 new_fluid_midi_router = cfunc('new_fluid_midi_router', POINTER(fluid_midi_router_t),
                                ('settings', c_void_p, 1),
-                               ('handler', CFUNCTYPE(POINTER(c_int), c_void_p, c_void_p), 1),
+                               ('handler', CFUNCTYPE(c_int, c_void_p, c_void_p), 1),
                                ('event_handler_data', c_void_p, 1))
 
-fluid_midi_router_handle_midi_event = cfunc('fluid_midi_router_handle_midi_event', POINTER(c_int),
+fluid_midi_router_handle_midi_event = cfunc('fluid_midi_router_handle_midi_event', c_int,
                                ('data', c_void_p, 1),
                                ('event', c_void_p, 1))
 
@@ -468,7 +568,7 @@ fluid_synth_get_sfont_by_id = cfunc('fluid_synth_get_sfont_by_id', c_void_p,
 fluid_sfont_get_preset = cfunc('fluid_sfont_get_preset', c_void_p,
                                ('sfont', c_void_p, 1),
                                ('banknum', c_int, 1),
-                               ('prognum', c_int, 1))
+                               ('prenum', c_int, 1))
 
 fluid_preset_get_name = cfunc('fluid_preset_get_name', c_char_p,
                               ('preset', c_void_p, 1))
@@ -520,14 +620,13 @@ class Synth:
         samplerate : output samplerate in Hz, default is 44100 Hz
         added capability for passing arbitrary fluid settings using args
         """
-        st = new_fluid_settings()
-        fluid_settings_setnum(st, b'synth.gain', gain)
-        fluid_settings_setnum(st, b'synth.sample-rate', samplerate)
-        fluid_settings_setint(st, b'synth.midi-channels', channels)
+        self.settings = new_fluid_settings()
+        self.setting('synth.gain', gain)
+        self.setting('synth.sample-rate', samplerate)
+        self.setting('synth.midi-channels', channels)
         for opt,val in kwargs.items():
             self.setting(opt, val)
-        self.settings = st
-        self.synth = new_fluid_synth(st)
+        self.synth = new_fluid_synth(self.settings)
         self.audio_driver = None
         self.midi_driver = None
         self.router = None
@@ -542,16 +641,17 @@ class Synth:
     def get_setting(self, opt):
         """get current value of an arbitrary synth setting"""
         val = c_int()
-        if fluid_settings_getint(self.settings, opt.encode(), byref(val)) != FLUID_FAILED:
+        if fluid_settings_getint(self.settings, opt.encode(), byref(val)) == FLUIDSETTING_EXISTS:
             return val.value
         strval = create_string_buffer(32)
-        if fluid_settings_copystr(self.settings, opt.encode(), strval, 32) != FLUID_FAILED:
+        if fluid_settings_copystr(self.settings, opt.encode(), strval, 32) == FLUIDSETTING_EXISTS:
             return strval.value.decode()
         num = c_double()
-        if fluid_settings_getnum(self.settings, opt.encode(), byref(num)) != FLUID_FAILED:
+        if fluid_settings_getnum(self.settings, opt.encode(), byref(num)) == FLUIDSETTING_EXISTS:
             return round(num.value, 6)
         return None
-    def start(self, driver=None, device=None, midi_driver=None):
+    
+    def start(self, driver=None, device=None, midi_driver=None, midi_router=None):
         """Start audio output driver in separate background thread
 
         Call this function any time after creating the Synth object.
@@ -567,7 +667,7 @@ class Synth:
         driver = driver or self.get_setting('audio.driver')
         device = device or self.get_setting('audio.%s.device' % driver)
         midi_driver = midi_driver or self.get_setting('midi.driver')
-        
+
         self.setting('audio.driver', driver)
         self.setting('audio.%s.device' % driver, device)
         self.audio_driver = new_fluid_audio_driver(self.settings, self.synth)
@@ -577,7 +677,14 @@ class Synth:
             new_fluid_cmd_handler(self.synth, self.router)
         else:
             fluid_synth_set_midi_router(self.synth, self.router)
-        self.midi_driver = new_fluid_midi_driver(self.settings, fluid_midi_router_handle_midi_event, self.router)
+        if midi_router == None: ## Use fluidsynth to create a MIDI event handler
+            self.midi_driver = new_fluid_midi_driver(self.settings, fluid_midi_router_handle_midi_event, self.router)
+            self.custom_router_callback = None
+        else:                   ## Supply an external MIDI event handler    
+            self.custom_router_callback = CFUNCTYPE(c_int, c_void_p, c_void_p)(midi_router)
+            self.midi_driver = new_fluid_midi_driver(self.settings, self.custom_router_callback, self.router)
+        return FLUID_OK
+
     def delete(self):
         if self.audio_driver:
             delete_fluid_audio_driver(self.audio_driver)
@@ -602,30 +709,30 @@ class Synth:
             fluid_synth_get_channel_info(self.synth, chan, byref(info))
             return (info.sfont_id, info.bank, info.program, info.name)
         else:
-            (sfontid, banknum, prognum) = self.program_info(chan)
-            presetname = self.sfpreset_name(sfontid, banknum, prognum)
-            return (sfontid, banknum, prognum, presetname)
+            (sfontid, banknum, presetnum) = self.program_info(chan)
+            presetname = self.sfpreset_name(sfontid, banknum, presetnum)
+            return (sfontid, banknum, presetnum, presetname)
     def program_info(self, chan):
         """get active soundfont, bank, prog on a channel"""
         if fluid_synth_get_program is not None:
             sfontid=c_int()
             banknum=c_int()
-            prognum=c_int()
-            fluid_synth_get_program(self.synth, chan, byref(sfontid), byref(banknum), byref(prognum))
-            return (sfontid.value, banknum.value, prognum.value)
+            presetnum=c_int()
+            fluid_synth_get_program(self.synth, chan, byref(sfontid), byref(banknum), byref(presetnum))
+            return (sfontid.value, banknum.value, presetnum.value)
         else:
             (sfontid, banknum, prognum, presetname) = self.channel_info(chan)
             return (sfontid, banknum, prognum)
-    def sfpreset_name(self, sfid, bank, prog):
+    def sfpreset_name(self, sfid, bank, prenum):
         """Return name of a soundfont preset"""
         if fluid_synth_get_sfont_by_id is not None:
             sfont=fluid_synth_get_sfont_by_id(self.synth, sfid)
-            preset=fluid_sfont_get_preset(sfont, bank, prog)
+            preset=fluid_sfont_get_preset(sfont, bank, prenum)
             if not preset:
                 return None
             return fluid_preset_get_name(preset).decode('ascii')
         else:
-            (sfontid, banknum, prognum, presetname) = self.channel_info(chan)
+            (sfontid, banknum, presetnum, presetname) = self.channel_info(chan)
             return presetname
     def router_clear(self):
         if self.router is not None:
@@ -781,16 +888,16 @@ class Synth:
         return fluid_synth_get_chorus_type(self.synth)
     def noteon(self, chan, key, vel):
         """Play a note"""
-        if key < 0 or key > 128:
+        if key < 0 or key > 127:
             return False
         if chan < 0:
             return False
-        if vel < 0 or vel > 128:
+        if vel < 0 or vel > 127:
             return False
         return fluid_synth_noteon(self.synth, chan, key, vel)
     def noteoff(self, chan, key):
         """Stop a note"""
-        if key < 0 or key > 128:
+        if key < 0 or key > 127:
             return False
         if chan < 0:
             return False
@@ -848,7 +955,46 @@ class Synth:
 
         """
         return fluid_synth_write_s16_stereo(self.synth, len)
+    def tuning_dump(self, bank, prog, pitch):
+        return fluid_synth_tuning_dump(self.synth, bank, prog, name.encode(), length(name), pitch)
 
+    def midi_event_get_type(self, event):
+        return fluid_midi_event_get_type(event)
+    def midi_event_get_velocity(self, event):
+        return fluid_midi_event_get_velocity(event)
+    def midi_event_get_key(self, event):
+        return fluid_midi_event_get_key(event)
+    def midi_event_get_channel(self, event):
+        return fluid_midi_event_get_channel(event)
+    def midi_event_get_control(self, event):
+        return fluid_midi_event_get_control(event)
+    def midi_event_get_program(self, event):
+        return fluid_midi_event_get_program(event)
+    def midi_event_get_value(self, event):
+        return fluid_midi_event_get_value(event)
+
+    def play_midi_file(self, filename):
+        self.player = new_fluid_player(self.synth)
+        if self.player == None: return FLUID_FAILED
+        if self.custom_router_callback != None:
+            fluid_player_set_playback_callback(self.player, self.custom_router_callback, self.synth)
+        status = fluid_player_add(self.player, filename.encode())
+        if status == FLUID_FAILED: return status
+        status = fluid_player_play(self.player)
+        return status
+    
+    def play_midi_stop(self):
+        status = fluid_player_stop(self.player)
+        if status == FLUID_FAILED: return status
+        status = fluid_player_seek(self.player, 0)
+        delete_fluid_player(self.player)
+        return status
+
+    def player_set_tempo(self, tempo_type, tempo):
+        return fluid_player_set_tempo(self.player, tempo_type, tempo)    
+
+
+    
 class Sequencer:
     def __init__(self, time_scale=1000, use_system_timer=True):
         """Create new sequencer object to control and schedule timing of midi events
